import pandas as pd
import plotly.express as px
import plotly.io as pio
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.offline import iplot
from sklearn.linear_model import LinearRegression
from typing import List
import numpy as np

class BaseVisualizer():
    """
    Creates graphs that illustrate general patterns that describe the entire dataset and 
    all of its users. 
    """
    def __init__(self, processed_df):
        self.df = processed_df

    def hist_feature_visualizer(self, column : str):

        """
        Creates histogram graphs for a specified column
        :param str column: column to be graphed
        """

        fig = px.histogram(self.df, x=column)
        iplot(fig)

    def create_visuals(self, features : List[str]):
        
        """
        Creates histograms for all of the specified columns
        :param list features: columns to be graphed
        """

        for f in features:
            if "id" not in f:
                self.hist_feature_visualizer(f)


import numpy as np
from sklearn.linear_model import LinearRegression
from statsmodels.tsa.ar_model import AutoReg
import math
from scipy.spatial import distance
import pandas as pd
from typing import List
from tqdm import tqdm 
tqdm.pandas()

def get_time_series_slope(time_series : List[int]) -> float:

    """
    Fits a line of best fit y=mx+b to the time series data
    and then extracts and returns the slope. Reveals the general 
    trend of the time series.
    """

    x_values = np.array(list(range(1, len(time_series)+1)))

    lr = LinearRegression()

    lr.fit(x_values.reshape(-1, 1), time_series)
    
    return lr.coef_[0]

def get_ar_params(time_series: List[int]) -> float:

    """
    Fits an autoregressive model to the time series
    and extracts it's parameters according to the formula
    y_t=δ+ϕ_1y_(t−1)+…+ϕ_py_(t−p)+ϵ_t.
    """

    model = AutoReg(time_series, lags=1)

    trained_model = model.fit()

    return trained_model.params


def get_shannon_entropy(time_series: List[int]) -> float:

    """
    Calculates the shannon entropy for a given time series
    Quantifies the amount of information contained within a variable
    """

    entropy_data = [x/sum(time_series) for x in time_series]

    s_entropy = 0

    for p in entropy_data:
        if p > 0:
            s_entropy += p * math.log(p, 2)

    return -s_entropy

def get_shannon_jensen_div(time_series: List[int]) -> float:

    """
    Calculates the shannon jensen divergence between the time series 
    scaled down between 0 and 1 compared to a uniform distribution.
    """

    time_series = np.array([x/sum(time_series) for x in time_series])

    uniform_dist = np.array([1/len(time_series) for x in time_series])

    return distance.jensenshannon(time_series, uniform_dist)

def get_num_changes(time_series: List[int]) -> float:

    """
    Counts the number of times a given time series experiences 
    changes from increasing to decreasing or decreasing to increasing.
    For example, the time series [1, 3, 2, 4, 1] -> 3
    """

    diffs = [time_series[i+1] - time_series[i] for i in range(len(time_series)-1)]
    extreme_count = 0

    for i in range(len(diffs)-1):
        if (diffs[i] < 0 and diffs[i+1] < 0) or (diffs[i] > 0 and diffs[i+1] > 0):
            pass
        else:
            extreme_count += 1

    return extreme_count


def get_avg_recurrence(time_series: List[int]) -> float:

    """
    Calculates the average length of each typing session 
    by determining the average length of time slices in which
    at least one keystroke event was made
    """

    total = 0
    count = 0
    values = []

    for idx in range(0, len(time_series)):
        if time_series[idx] == 0:
            count += 1
            values.append(total)
            total = 0
        else:
            total += 1
        
    if (time_series[len(time_series)-1] != 0):
        values.append(total)
        count += 1
            
    return sum(values) / count

def get_stddev_recurrence(time_series: List[int]) -> float:

    """
    Calculates the standard deviation of the length of typing sessions
    in which at least one keystroke event was made
    """

    total = 0
    count = 0
    values = []

    for idx in range(0, len(time_series)):
        if time_series[idx] == 0:
            count += 1
            values.append(total)
            total = 0
        else:
            total += 1
        
    if (time_series[len(time_series)-1] != 0):
        values.append(total)
        count += 1
            
    return np.std(values)

def get_cursor_back(time_series: List[int]) -> int:

    """
    Counts the number of times the cursor went backwards
    """

    return len([x for x in time_series if x < 0])

def count_bursts(data: List[int], min_burst_val=0.01) -> int:

    """
    Measures how many periods of times there were where 
    the difference between keystroke events were less than 
    the min_burst_val. Also measures the average length of the 
    pauses between bursts
    """

    burst_count = 0

    burst_start = 0

    pause_total = 0

    pause_count = 0

    bursts = []

    for i in range(1, len(data)):
        if data[i] > min_burst_val:
            pause_total += data[i]
            pause_count += 1
            if i - burst_start - 1 > 0:
                burst_count += 1
                bursts.append(data[i])

            burst_start = i

    if (burst_start != len(data)-1):
        burst_count += 1

    return [burst_count, pause_total/pause_count]

class DataHandler():
    """
    Main data processor class. Expects a dataframe with the following columns:  

    Required:
    id: the session of writing a particular keystroke belongs to
    event id: the order in which a particular keystroke occured relative to the typing session
    down time/up time: the time when a key was pressed and released

    Recommended:
    event: the type of event occuring for a keystroke (if data is anonymized, then any key press defaults to some filler letter e.g. q) 
    cursor_position: the location of the cursor
    """

    def __init__(self, df, config, target_column=None):
        self.df = df
        self.target_column = target_column
        self.config = config

        if "id" not in self.config.features or "event_id" not in self.config.features or "down_time" not in self.config.features:
            raise ValueError("Must have all three required features ('id', 'event_id', 'down_time')")
    
    def feature_engineer(self) -> pd.DataFrame:

        """
        For a raw keystroke log dataframe, we create a new dataframe suitable for supervised learning tasks.
        Feature Descriptions:
        Pause value(s): The number of pauses during the writing session that by duration were in between the value specified and the next
        value specified
        Mean pause duration: the average length of pauses between bursts in seconds
        Burst count: the number of bursts during the typing session
        Verbosity: total number of keystrokes in the typing session
        Backspace count: total number of backspaces occuring during the typing session
        Word count: total number of words typed out during the typing session
        Sent count: total number of sentences during the typing session
        Paragraph count: total number of paragraphs during the typing session
        Nonproduction: total number of keystrokes not contributing to the overall text
        Average keystroke speed: average amount of time spent on each keystroke
        AR P1/P2: The two parameters of an autoregressive model fit to the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Slope degree: the slope of a linear model fit to the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Entropy: the shannon entropy of the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Degree uniformity: the shannon-jensen divergence of the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Local extremes: the number of local highs/lows occuring in the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Average recurrence: the average length of consecutive time slices where at least one keystroke event occured
        Stddev recurrence: the standard deviation of lengths of consecutive time slices where at least one keystroke event occured
        Largest latency: the biggest difference in time between two keystrokes
        Median Latency: the median of differences between keystrokes
        Smallest latency: the smallest difference between keystrokes
        First pause: the initial pause at the start of the typing session
        Cursor Back: the number of times during the typing session where the cursor moved backwards
        Word Back: the number of words that were deleted during the typing session
        Largest insert: the greatest number of words pasted into the text
        Largest delete: the greatest number of words removed from the text
        """

        self.df.sort_values(['id', 'up_time'], inplace=True)

        self.df['diffs'] = self.df.groupby(['id'])['up_time'].transform(lambda x: x.diff())

        self.df.sort_index(inplace=True)
        self.df["diffs_seconds"] = self.df["diffs"] / 1000

        self.df["time_elapsed"] = self.df.groupby("id")["diffs_seconds"].cumsum()
        self.df["time_elapsed"].fillna(0, inplace=True)

        new_df = pd.DataFrame({"id" : list(self.df.groupby("id").groups.keys())})

        for i in range(len(self.config.pause_vals[:-1])):
            pause_counts = self.df[(self.df["diffs_seconds"] > self.config.pause_vals[i]) & (self.df["diffs_seconds"] < self.config.pause_vals[i+1])].groupby("id").size()
            for id_ in new_df["id"].values:
                if pause_counts.get(id_) is None:
                    pause_counts[id_] = 0

            new_df[f"pause_{self.config.pause_vals[i]}"] = pause_counts.values

        pause_counts = self.df[self.df["diffs_seconds"] > self.config.pause_vals[-1]].groupby("id").size()

        for id_ in new_df["id"].values:
            if pause_counts.get(id_) is None:
                pause_counts[id_] = 0

        new_df[f"pause_{self.config.pause_vals[-1]}"] = pause_counts.values

        df_grouped = self.df.groupby("id")

        burst_vals = np.array(df_grouped.progress_apply(lambda x: count_bursts(x["diffs_seconds"].values)).values.tolist())

        new_df["mean_pause_duration"] = burst_vals[:,1]
        new_df["burst_count"] = burst_vals[:,0]

        new_df["verbosity"] = df_grouped.size().values

        if "up_event" in self.config.features:
            backspace_df = self.df.groupby(["up_event", "id"]).size()["Backspace"]
            new_df = pd.merge(new_df, backspace_df.rename("backspaces"), on="id", how="left")

            new_df["backspaces"].fillna(0, inplace=True)

            period_df = self.df.groupby(["up_event", "id"]).size()["."]
            new_df = pd.merge(new_df, period_df.rename("sent_count"), on="id", how="left")

            enter_df = self.df.groupby(["up_event", "id"]).size()["Enter"]
            new_df = pd.merge(new_df, enter_df.rename("paragraph_count"), on="id", how="left")


        if "word_count" in self.config.features:
            new_df["word_count"] = df_grouped["word_count"].last().values

        if "activity" in self.config.features:
            nonprod_df = self.df.groupby(["activity", "id"]).size()["Nonproduction"]
            new_df = pd.merge(new_df, nonprod_df.rename("Nonproduction"), on="id", how="left")

        new_df["avg_keystroke_speed"] = new_df["verbosity"] / df_grouped["time_elapsed"].tail(1).values

        for window in self.config.window_values:
            self.df[f"window_{window}_sec_idx"] = self.df["time_elapsed"].apply(lambda x: math.floor(x / window))
            ar = np.array(df_grouped.progress_apply(lambda x: get_ar_params(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values.tolist())
            ar_p1 = ar[:,0]
            ar_p2 = ar[:,1]
            new_df[f"ar_{window}_p1"] = ar_p1
            new_df[f"ar_{window}_p2"] = ar_p2

            new_df[f"Slope_Degree_{window}"] = df_grouped.progress_apply(lambda x: get_time_series_slope(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values
            new_df[f"Entropy_{window}"] = df_grouped.progress_apply(lambda x: get_shannon_entropy(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values
            new_df[f"Degree_Uniformity_{window}"] = df_grouped.progress_apply(lambda x: get_shannon_jensen_div(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values
            new_df[f"Local_Extremes_{window}"] = df_grouped.progress_apply(lambda x: get_num_changes(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values
            new_df[f"Average_Recurrence_{window}"] = df_grouped.progress_apply(lambda x: get_avg_recurrence(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values
            new_df[f"StdDev_Recurrence_{window}"] = df_grouped.progress_apply(lambda x: get_stddev_recurrence(x[f"window_{window}_sec_idx"].value_counts().reindex(range(max(x[f"window_{window}_sec_idx"])+1), fill_value=0))).values

        new_df["largest_latency"] = df_grouped["diffs"].max().values

        new_df["smallest_latency"] = df_grouped["diffs"].min().values

        new_df["median_latency"] = df_grouped["diffs"].median().values

        new_df["first_pause"] = self.df.groupby("id").diffs_seconds.first().values

        if "cursor_position" in self.config.features:
            self.df['curpos_diffs'] = self.df.groupby(['id'])['cursor_position'].transform(lambda x: x.diff())
            new_df["Cursor_Back_Count"] = df_grouped.progress_apply(lambda x: get_cursor_back(x["curpos_diffs"])).values

        if "word_count" in self.config.features:
            self.df['word_diffs'] = self.df.groupby(['id'])['word_count'].transform(lambda x: x.diff())
            new_df["Word_Back_Count"] = df_grouped.progress_apply(lambda x: get_cursor_back(x["word_diffs"])).values
            new_df["largest_insert"] = df_grouped["word_diffs"].max().values
            new_df["largest_delete"] = df_grouped["word_diffs"].min().values

        return new_df 
    
    def get_df(self):
        return self.feature_engineer()


from visuals.temporal_analysis import TemporalVisualizer
from visuals.base_analysis import BaseVisualizer
from utils.config import KeystrokeConfig
from utils.data_processing import DataHandler
from lightgbm import LGBMRegressor
from sklearn.linear_model import LinearRegression
import pandas as pd
import numpy as np
import pickle
import json
import os

class KeystrokeTrainer():

    """
    Main class for keystroke modeling and analysis. Requires the path
    to the raw keystroke data. Additionally, one can pass in a custom configuration
    if different settings are desired. 
    """

    def __init__(self, data_pth : str, config : KeystrokeConfig = KeystrokeConfig()):
        self.raw_df = pd.read_csv(data_pth)
        self.columns = self.raw_df.columns
        self.config = config
        
        if "id" not in self.columns or "event_id" not in self.columns or "down_time" not in self.columns:
            raise ValueError("Must have all three required features ('id', 'event_id', 'down_time') in dataframe. Please rename these columns before passing in the csv path")
        
        print("-"*10 + "Processing Data" + "-"*10 + "\n")

        self.data_handler = DataHandler(self.raw_df, self.config)
        self.processed_df = self.data_handler.get_df()

        print(f"{len(self.processed_df)} samples of data created with {self.processed_df.shape[1]} features\n")

        print("-"*10 + "Feature Descriptions" + "-"*10 + "\n")
        print("""
        Pause value(s): The number of pauses during the writing session that by duration were in between the value specified and the next
        value specified
        Mean pause duration: the average length of pauses between bursts in seconds
        Burst count: the number of bursts during the typing session
        Verbosity: total number of keystrokes in the typing session
        Backspace count: total number of backspaces occuring during the typing session
        Word count: total number of words typed out during the typing session
        Sent count: total number of sentences during the typing session
        Paragraph count: total number of paragraphs during the typing session
        Nonproduction: total number of keystrokes not contributing to the overall text
        Average keystroke speed: average amount of time spent on each keystroke
        AR P1/P2: The two parameters of an autoregressive model fit to the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Slope degree: the slope of a linear model fit to the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Entropy: the shannon entropy of the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Degree uniformity: the shannon-jensen divergence of the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Local extremes: the number of local highs/lows occuring in the number of keystrokes occuring over time within varying time intervals 
        controlled by the user in the config
        Average recurrence: the average length of consecutive time slices where at least one keystroke event occured
        Stddev recurrence: the standard deviation of lengths of consecutive time slices where at least one keystroke event occured
        Largest latency: the biggest difference in time between two keystrokes
        Median Latency: the median of differences between keystrokes
        Smallest latency: the smallest difference between keystrokes
        First pause: the initial pause at the start of the typing session
        Cursor Back: the number of times during the typing session where the cursor moved backwards
        Word Back: the number of words that were deleted during the typing session
        Largest insert: the greatest number of words pasted into the text
        Largest delete: the greatest number of words removed from the text
        """)

        if self.config.target_col is not None:
            target_values = pd.read_csv(self.config.target_path)
            self.processed_df = pd.merge(self.processed_df, target_values, on="id")

        self.temporal_visualizer = TemporalVisualizer(self.raw_df)
        self.base_visualizer = BaseVisualizer(self.processed_df)

        self.results = {"results" : []}

    def train(self, estimator, train_df : pd.DataFrame, targets : pd.Series):
        """
        Trains an estimator on the processed data and returns the trained
        estimator instance. Expects a model with scikit-learn syntax (LightGBM,
        XGBoost, CatBoost supported)
        """
        if not os.path.exists(f"{self.config.output_dir}"):
            os.makedirs(f"{self.config.output_dir}")

        if self.config.validate:

            from sklearn.model_selection import train_test_split
            X_train, X_valid, y_train, y_valid = train_test_split(train_df, targets.values, test_size=self.config.split_size, random_state=2007, shuffle=True)

            estimator.fit(X_train, y_train)

            with open(f"{os.path.abspath(self.config.output_dir + estimator.__class__.__name__)}.pkl", "wb") as f:
                pickle.dump(estimator, f)

            valid_pred = estimator.predict(X_valid)

            if self.config.task == "reg":
                from sklearn.metrics import r2_score, mean_squared_error
                self.results["results"].append(
                    {
                        "Model Name" : estimator.__class__.__name__,
                        "Metrics" : 
                                [
                                    ("Mean Squared Error", mean_squared_error(y_valid, valid_pred)),
                                    ("R2 Score", r2_score(y_valid, valid_pred)),
                                ], 
                    }
                )
            
            elif self.config.task == "cls":

                from sklearn.metrics import log_loss, accuracy_score

                valid_proba = estimator.predict_proba(X_valid)

                self.results["results"].append(
                    {
                        "Model Name" : estimator.__class__.__name__,
                        "Metrics" : 
                                [
                                    ("Log Loss", log_loss(y_valid, valid_proba)),
                                    ("Accuracy", accuracy_score(y_valid, valid_pred)),
                                ]
                    }
                )       
        else:
            estimator.fit(train_df, targets.values)
            with open(f"{self.config.output_dir}{estimator.__class__.__name__}.pkl") as f:
                pickle.dump(estimator, f)

    def run(self):

        print("-"*10 + "Performing Data Analysis" + "-"*10 + "\n")

        print(f"{2 * self.config.user_samples + len(self.processed_df.columns)} graphs created\n")

        self.temporal_visualizer.create_user_visuals(self.config.user_samples)
        self.base_visualizer.create_visuals(self.processed_df.columns)

        if self.config.target_col is None:

            print("*"*20, "Keystroke graphics provided\nNo modeling performed as the target column was None", "*"*20)
        
        else:

            targets = self.processed_df.pop(self.config.target_col)

            for estimator in self.config.models:
                self.train(estimator, self.processed_df.drop("id", axis=1), targets)

            for result in self.results["results"]:
                print("\n\n")
                print("-"*10 + f"{result['Model Name']} Training Results" + "-"*10 + "\n")
                for metric in result["Metrics"]:
                    print(f"{metric[0]} : {metric[1]}\n")

        
        with open(f'{self.config.output_dir}/results.json', 'w') as fp:
            json.dump(self.results, fp)
    



from lightgbm import LGBMRegressor

class KeystrokeConfig():

    """
    Configuration class for data preparation and training. 
    """

    def __init__(self, 
                 min_burst_val=0.01,
                 window_values=[30.0, 60.0],
                 features=["id", "event_id", "down_time", "word_count", "cursor_position"],
                 pause_vals=[0.5, 1, 1.5, 2, 3],
                 user_samples=3,
                 target_col=None,
                 validate=False,
                 split_size=None,
                 task="reg",
                 models=[LGBMRegressor()],
                 output_dir="output/",
                 target_path=None,
                 ):
        self.min_burst_val = min_burst_val
        self.window_values = window_values
        self.features = features
        self.pause_vals = pause_vals
        self.user_samples = user_samples
        self.target_col = target_col
        self.validate = validate
        self.models = models
        self.task = task 
        self.split_size = split_size
        self.output_dir = output_dir
        self.target_path = target_path


import pandas as pd
import plotly.express as px
import plotly.io as pio
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.offline import iplot
from sklearn.linear_model import LinearRegression
from typing import List
import numpy as np
import random
import math

def view_keystrokes_per_time(num_keystrokes : List[int], width=1800, height=700, user_id=""):

    """
    Creates a graph of the number of keystrokes occuring in a certain slice of time for all slices
    in the typing session. Additionally, a line of best fit is plotted so we can see the general trend.

    :param list num_keystrokes: The series of keystrokes per time slice
    :param int width: the width of the graph
    :param int height: the height of the graph
    """

    x_vals = list(range(1, len(num_keystrokes)+1))
    x_vals = np.array(x_vals).reshape(-1, 1)

    lr = LinearRegression()
    lr.fit(x_vals, num_keystrokes)
    best_fit = lr.predict(x_vals)
        
    fig = make_subplots()

    trace1 = go.Scatter(
        x=x_vals.squeeze(),
        y=num_keystrokes,
        name='# Keystrokes Per Time Slice',
    )

    trace2 = go.Line(
        x=x_vals.squeeze(),
        y=best_fit,
        name="Best Fit"
    )

    fig.add_trace(trace1)
    fig.add_trace(trace2)  

    fig.update_layout(
            title=f"Keystroke Verbosity Per Time Slice for User {user_id}",
            xaxis_title="Time Slice",
            yaxis_title="Frequency",
            width=width,
            height=height,
        )

    iplot(fig)

def view_time_per_keystroke(keystroke_diffs : List[float], width=1800, height=700, user_id=""):

    """
    Creates a graph of the time differences between keystrokes in a typing session. 

    :param list keystroke_diffs: The series of differences between subsequent keystroke events
    :param int width: the width of the graph
    :param int height: the height of the graph
    """

    fig = make_subplots()
  
    trace1 = go.Scatter(
        x=list(range(1, len(keystroke_diffs) + 1)),
        y=keystroke_diffs,
        name='Time Spent Between Next Keystroke',
    )

    fig.add_trace(trace1)

    fig.update_layout(
            title=f"Time Spent Between Keystroke for User {user_id}",
            xaxis_title="Keystroke Event Number",
            yaxis_title="Time (seconds)",
            width=width,
            height=height,
        )

    iplot(fig)

class TemporalVisualizer():

    """
    Primary class for visualizing keystroke analytics
    This class creates graphics for time based relationships
    """

    def __init__(self, keystroke_df):
        self.df = keystroke_df
        self.df.sort_values(['id', 'up_time'], inplace=True)

        self.df['diffs'] = self.df.groupby(['id'])['up_time'].transform(lambda x: x.diff())

        self.df.sort_index(inplace=True)
        self.df["diffs_seconds"] = self.df["diffs"] / 1000

        self.df["time_elapsed"] = self.df.groupby("id")["diffs_seconds"].cumsum()
        self.df["time_elapsed"].fillna(0, inplace=True)

    def _sample_user(self) -> pd.DataFrame:

        """
        Samples data from a particular users typing session
        """

        rand_idx = random.randint(0, len(self.df["id"].unique())-1)
        return self.df[self.df["id"] == self.df["id"].unique()[rand_idx]]
    
    def create_user_visuals(self, num_users : int = 1):

        """
        Creates visuals for a particular users typing session

        :param int num_users: number of users to sample and visualize
        """

        for i in range(num_users):
            user_data = self._sample_user()
            user_data[f"window_30_sec_idx"] = user_data["time_elapsed"].apply(lambda x: math.floor(x / 30))
            view_keystrokes_per_time(user_data.groupby("window_30_sec_idx").size().values, user_id=user_data["id"].values[0])
            view_time_per_keystroke(user_data["diffs_seconds"].values, user_id=user_data["id"].values[0])